# coding: utf-8
import json
import logging
import os
from typing import Dict, Optional, Union

from rosreestr2coord.request.exceptions import HTTPForbiddenException

from .export import coords2kml
from .logger import logger
from .request.proxy_handling import ProxyHandling
from .request.request import make_request
from .utils import clear_code, code_to_filename

TYPES = {
    "Участки": (1, "RR:PARCEL_V"),
    # "ОКС": 5,
    "Кварталы": (2, "RR:CAD_BLOCK"),
    # "Районы": 3,
    # "Округа": 4,
    # "Границы": 7,
    # "ЗОУИТ": 10,
    # "Тер. зоны": 6,
    # "Красные линии": 13,
    # "Лес": 12,
    # "СРЗУ": 15,
    # "ОЭЗ": 16,
    # "ГОК": 9,
}


class NoCoordinatesException(Exception):
    pass


class Area:

    def __init__(
        self,
        code: str = "",
        area_type: int = 1,
        media_path: str = "",
        with_log: bool = True,
        coord_out: str = "EPSG:4326",
        with_proxy: bool = False,
        use_cache: bool = True,
        proxy_handler: Optional[ProxyHandling] = None,
        timeout: int = 5,
        logger: Optional[logging.Logger] = logger,
        proxy_url: Optional[str] = None,
    ):
        self.with_log = with_log
        self.area_type = self._get_code_by_number(area_type)
        self.media_path = media_path

        self.code = code

        self.proxy_url = proxy_url

        self.file_name = code_to_filename(self.code[:])
        self.with_proxy = with_proxy
        self.proxy_handler = proxy_handler
        self.logger = logger
        self.use_cache = use_cache
        self.coord_out = coord_out
        self.timeout = timeout

        self.session_id = None
        self.max_session_id_attempt = 5
        self.session_id_attempt = 0

        self.feature = None

        if not code:
            return
        self.tmp_path = self.create_tmp()
        self.workspace = self.create_workspace()
        self._update_session_id()
        try:
            geom = self.get_geometry()
            if not geom:
                self.log("Nothing found")
        except Exception as er:
            if hasattr(er, "reason"):
                self.log(er.reason)
            else:
                self.log(er)

    def create_tmp(self):
        if not self.media_path:
            self.media_path = os.getcwd()
        tmp_path = os.path.join(self.media_path, "tmp")
        if not os.path.isdir(tmp_path):
            os.makedirs(tmp_path)
        return tmp_path

    def create_workspace(self):
        area_path_name = code_to_filename(clear_code(self.code))
        workspace = os.path.join(self.tmp_path, area_path_name)
        if not os.path.isdir(workspace):
            os.makedirs(workspace)
        return workspace

    @staticmethod
    def _get_code_by_number(number: int) -> str:
        for key, value in TYPES.items():
            if value[0] == number:
                return value[1]
        return None

    def search(self):
        url = "https://nspd.gov.ru/map_api/s_search/search"
        payload = {
            "text": self.code,
            "pageNumber": 0,
            "pageSize": 1,
        }
        resp = self.make_request(url, method="POST", body=payload)
        response_str = resp.decode("utf-8")
        response_obj = json.loads(response_str)
        return response_obj["result"][0]

    def get_geometry(self):

        cad_item = self.search()

        url = "https://nspd.gov.ru/geoserver/wfs"
        params = [
            "service=wfs",
            "version=2.0.0",
            "OUTPUTFORMAT=application/json",
            "REQUEST=GetFeature",
            f"TYPENAME={cad_item["className"]}",
            f"cql_filter=cad_num='{self.code}'",
            f"authkey={self.session_id}",
            f"srsName={self.coord_out}",
        ]
        url = f"{url}?{'&'.join(params)}"

        resp = self.make_request(url)
        response_str = resp.decode("utf-8")
        response_obj = json.loads(response_str)
        feature = response_obj["features"][0]
        self.feature = feature
        return feature

    # Deprecated use to_geojson for all cases
    def to_geojson_poly(self, dumps=True):
        return self.to_geojson(dumps)

    def to_geojson(self, dumps=True):
        feature_collection = self.feature
        if feature_collection:
            if dumps:
                return json.dumps(feature_collection)
            return feature_collection
        return False

    def to_kml(self):
        coords = [self.feature["geometry"]["coordinates"]]
        attrs = self.feature["properties"]
        return coords2kml(coords, attrs)

    def _update_session_id(self):
        if not self.session_id:
            url = "https://nspd.gov.ru/map_api/auth/authorize"
            resp = self._make_request(url=url, method="POST", body={"userId": None, "roles": []})
            resp_str = resp.decode("utf-8")
            resp_json = json.loads(resp_str)

            self.session_id = resp_json["sessionId"]

    def _make_request(
        self,
        url: str,
        method: str,
        body: Optional[Union[Dict, bytes]],
    ):
        proxy_path = os.path.join(self.tmp_path, "proxy.txt")
        proxy_handler = self.proxy_handler if self.proxy_handler else ProxyHandling(path=proxy_path)
        self.logger.debug(url)
        headers = {
            "Content-Type": "application/json",
        }
        response = make_request(
            url=url,
            body=body,
            method=method,
            with_proxy=self.with_proxy,
            proxy_handler=proxy_handler,
            logger=self.logger,
            timeout=self.timeout,
            headers=headers,
            proxy_url=self.proxy_url,
        )
        return response

    def make_request(
        self,
        url: str,
        method: str = "GET",
        body: Union[Dict, bytes, None] = None,
    ):
        self._update_session_id()
        try:
            resp = self._make_request(url, method, body)
            self.session_id_attempt = 0
            return resp
        except HTTPForbiddenException:
            if self.max_session_id_attempt < self.session_id_attempt:
                self.session_id = None
                self.session_id_attempt = self.session_id_attempt + 1
                return self.make_request(url, method=method, body=body)
            raise
        except Exception:
            raise

    @staticmethod
    def clear_code(code):
        return clear_code(code)

    def log(self, msg):
        if self.with_log:
            print(msg)

    def error(self, msg):
        if self.with_log:
            self.logger.warning(msg)
